"""
Basic API for active matter simulations
"""

import os
import numpy
import random
import logging
import csv

import argh
import numpy as np
from atooms.core.utils import setup_logging, mkdir, rmf
from atooms.simulation import Simulation
from atooms.trajectory import Trajectory, TrajectoryXYZ
from atooms.simulation.observers import write_trajectory, write, Scheduler

try:
    from vicsek import Vicsek
    from neighbors import VicsekNeighbors
except:
    from atooms.active.vicsek import Vicsek
    from atooms.active.neighbors import VicsekNeighbors

# Utilities


def _clear_logging():
    import logging
    log = logging.getLogger()
    for h in log.handlers[-2:]:
        log.removeHandler(h)

def _initialize_system(dim, L, npart, file_inp, load=False, save=False):
    """
    System initialization.
    The system can be generated on the spot, loaded from an input file or saved to disk
    """
    # TODO: add metadata read
    # TODO: move to utilities module
    # Import pkgs
    from atooms.system import System
    from atooms.system.cell import Cell

    # Set input trajectory file path
    input_file = os.path.join(os.path.dirname(__file__), file_inp)

    if (load == True and os.path.exists(input_file) == True):
        # Load system from file (file exists)
        logging.info("Load starting config from file")
        with TrajectoryXYZ(input_file, 'r') as th:
            system = th[0]
    elif (load == True and os.path.exists(input_file) == False):
        # Load system from file (file doesn't exist)
        raise Exception('Failed to load config from file.')
    elif not load:
        # Initialize system from scratch
        system = System(N=npart)
        system.cell = Cell(np.ones(dim)*L)
        # TODO: build system in place
        for p in system.particle:
            p.position = [L*np.random.uniform() for i in range(dim)]
            p.velocity = np.zeros(dim)
            p.orientation = np.random.uniform(-np.pi, np.pi)
            p.radius = None
            # apply PBC - slow, but it's just one call
            p.fold(system.cell)

    if (save and not load):
        # Save to file
        logging.info("Save starting config to file")
        logging.info("Existing start_traj.xyz will be overwritten")
        with TrajectoryXYZ(input_file, 'w') as th:
            th.variables = ['pos', 'orientation', 'velocity']
            th.write(system)
    elif (save and load):
        raise Exception("Cannot load and save system at the same time. Set load or save to False.")

    return system

def _random_choice_id(length=8, seed=None):
    """
    Generate UniqueIDentifier.
    RNG can be independent of global seed set by random.seed or numpy.random.seed
    """
    # TODO: move to utilities module (it can be useful elsewhere)
    import string
    rng = random.Random(seed)
    alphabet = string.ascii_lowercase + string.digits
    return ''.join(rng.choices(alphabet, k=length))

def _order_parameter(sim, v0):
    # TODO: get view
    velocity = sim.system.dump('velocity')
    norm = np.linalg.norm(np.sum(velocity, axis=1))/len(sim.system.particle)
    return norm/v0

# API


@argh.arg('--load', action='store_true')
@argh.arg('--save', action='store_true')
def vm(file_inp, file_out=None, load=False, save=False, noise='vectorial',
       neighbors='kdtree', dim=2, npart=2048, eta=0.4, rho=2., v0=1.,
       nsteps=1000, checkpoint_interval=0, config_number=0, verbose=False,
       seed=1, seed_inp='data/seed.txt', block='', restart=False, dry=False, 
       log=True, data_log='data/metadata.log'):
    """
    Simple API with logging facilities for Vicsek model simulations.

    Parameters
    ----------
    file_inp:               string
                            Input file path.
    file_out:               string, default None
                            Output file path. If None, output files are created in `data/`, using a randomly generated unique identifier as file name.
    load:                   bool, default False
                            Enable loading of starting trajectory from `file_inp`
    save:                   bool, default False
                            Enable saving of starting trajectory to `file_out`
    noise:                  string, default 'vectorial'
                            Orientation update method. See `active.vicsek` module.
    neighbors:              string, default `kdtree`
                            Fixed radius neighhbor search algorithm. Akin to `method` parameter in `active.neighbors`.
    dim:                    int, default 2
                            System dimensionality.
    npart:                  int, default 2048
                            Number of particles.
    eta:                    float, default 0.4
                            Noise amplitude
    rho:                    float, default 2.0
                            Particle number density.
    v0:                     float, default 1.0
                            Velocity module.
    nsteps:                 int, default 1000
                            Number of simulation steps.
    checkpoint_interval:    int, default 0
                            Save checkpoint trajectory at selected interval.
    config_number:          int, default 0
                            Number of trajectories to save.
    verbose:                bool, default False
                            Enable verbose logging.
    seed:                   int, default 1
                            Seed for the random number generator.
    seed_inp:               string, default 'data/seed.txt'
                            Seed file path. It has priority over `seed` variable.
    restart:                bool, default False
                            Enable simulation restarting from checkpoint.
    dry:                    bool, default False
                            Enable dry run. No simulation is performed.
    log:                    bool, default True
                            Enable system parameters logging in `data_log`.
    data_log:               string, default 'data/metadata.log'
                            System parameters logging path.
    """

    # Initialize seed and generate uid
    if os.path.exists(seed_inp):
        f = open(seed_inp, 'r')
        seed = int(f.read())
    random.seed(seed)
    numpy.random.seed(seed)
    uid = _random_choice_id(length=16)

    # Initialize filenames and directory structure
    if file_out == None:
        file_out = 'data/'+uid+'.xyz'
    mkdir(os.path.dirname(file_out))

    # Always log to file
    if not restart:
        rmf(file_out + '.log')
    setup_logging(level=20, filename=file_out + '.log')
    if verbose:
        setup_logging(level=20)

    # Build system or read initial state, depending on user action
    L = np.sqrt(npart/rho)
    system = _initialize_system(dim, L, npart, file_inp, load=load, save=save)

    # Simulation backend
    bck = Vicsek(system, eta, v0, noise=noise)

    # Neighbor search method
    # TODO: evaluate whether to keep or remove pseudo-factory VicsekNeighbors
    # by moving method selection over here
    neighs = VicsekNeighbors(system, method=neighbors)
    bck.neighbors = neighs

    # Simulation
    sim = Simulation(bck, output_path=file_out, steps=nsteps,
                     enable_speedometer=True, restart=restart,
                     checkpoint_interval=checkpoint_interval)
    sim.checkpoint_variables = ['position', 'orientation', 'velocity']

    # Trajectory class instantiation with proper metadata
    trajectory = TrajectoryXYZ(file_out, mode='w')
    trajectory.metadata = {'eta': eta, 'v0': v0, 'seed': seed, 'noise': noise, 'neighbors': neighbors}

    # Writing trajectory file
    if config_number == 0:
        if len(block) > 0:
            sim.add(write_trajectory, Scheduler(block=[int(_) for _ in block.split(',')]),
                    variables=['position', 'orientation', 'velocity'],
                    trajectory=trajectory)
    else:
        sim.add(write_trajectory, Scheduler(calls=config_number),
                variables=['position', 'orientation', 'velocity'],
                trajectory=trajectory)

    # Get order parameter
    # TODO: evaluate whether to move in postprocessing or to keep in API;
    # it's a useful measure of how "correct" the simulation is at a glance

    # TODO: fix write to file - cbk tuple doesn't work properly
    # cbk = ('phi', _order_parameter)
    # sim.add(write, Scheduler(calls=config_number), cbk, suffix='phi')

    # Create/update parameter log
    if log:
        with open(data_log, 'a') as f:
            write = csv.writer(f)
            # Add header if file is empty
            if os.stat(data_log).st_size == 0:
                write.writerow(['npart', 'rho', 'eta', 'v0', 'nsteps', 'noise', 'seed', 'uid'])
            write.writerow([npart, rho, eta, v0, nsteps, noise, seed, uid])

    # Run
    if not dry:
        sim.run()
        _clear_logging()

    return sim


if __name__ == '__main__':
    argh.dispatch_command(vm)



"""
Vicsek model simulation backend.
"""

import os
import numpy as np
from f2py_jit import jit

__all__ = ['Vicsek']

_f90 = jit("""
subroutine fold(pos, center, side)
	implicit none
	double precision, intent(inout) :: pos(:,:)
	double precision, intent(in)	:: center(:), side(:)
	integer							:: i
	do i = 1,size(pos,2)
		pos(:,i) = pos(:,i) - center
		pos(:,i) = pos(:,i) - nint(pos(:,i)/side)*side
		pos(:,i) = pos(:,i) + center
	end do
end subroutine fold
""", flags="-O3 -ffast-math")


# Backend
def noise_scalar(orientations, number_neighbors, sum_neighbors, eta, n):
    """Compute orientations with scalar noise (Vicsek et al, 1995; Grégoire and Chaté, 2008)."""
    rnd = eta * np.random.uniform(-np.pi, np.pi, size=n)
    avg = np.angle(sum_neighbors)
    do = avg+rnd
    return do

def noise_vectorial(orientations, number_neighbors, sum_neighbors, eta, n):
    """Compute orientations with vectorial noise (Grégoiré and Chatè, 2008)"""
    rnd = number_neighbors*eta * np.exp(1j*np.random.uniform(-np.pi, np.pi, size=n))
    do = np.angle(sum_neighbors+rnd)
    return do


_noise_db = {'scalar': noise_scalar,
             'vectorial': noise_vectorial}

class Vicsek(object):
    """
    Base Vicsek model simulation backend.

    Parameters
    ----------
    system: atooms system
                    `atooms` system to be evolved during the simulation.
    eta: 	float
                    Noise amplitude.
    v0: 	float
                    Velocity module.
    noise: 	{'scalar', 'vectorial'}, default 'vectorial'
                    Orientation update method. Available methods are `scalar`, which employs the scalar noise discussed in Vicsek et al, 1995; `vectorial` which employs the vectorial noise discussed in Grégoiré and Chatè, 2008.
    radius: float, default 1.0
                    Interaction radius.
    deltat: float, default 1.0
                    Time step.
    """

    def __init__(self, system, eta, v0, noise='vectorial', radius=1.0, deltat=1.0):
        self.system = system
        self.eta = eta
        self.v0 = v0
        self.deltat = deltat
        self.radius = radius
        # Setup neighbor related variables
        self.neighbors = None
        self.sum_neighbors = None
        self.number_neighbors = None

        # Set noise
        if noise in _noise_db:
            self.noise = _noise_db[noise]
        else:
            assert hasattr(noise, '__call__'), f'noise must be in {list(_noise_db.keys())} or callable'
            self.noise = noise

    def __str__(self):
        return f"""
		backend: Vicsek model
		algorithm: {self.noise}
		timestep: {self.deltat}
"""

    def _fold(self, method='f90'):
        """
        Fold self into central cell.
        View-based fork of atooms.system.particle folding routine
        """
        pos = self.system.view('pos', order='F')
        npart = len(self.system.particle)
        ndim = len(self.system.cell.side)
        if method == 'python':
            # Setup Fortran ordered arrays
            center = np.ndarray(shape=(ndim, npart), order='F', dtype=np.float64)
            center[:, :] = self.system.cell.center[:, None]
            side = np.ndarray(shape=(ndim, npart), order='F', dtype=np.float64)
            side[:, :] = self.system.cell.side[:, None]

            # Move the center to 0 when cell isn't centered in 0
            if not np.all(self.system.cell.center == 0.):
                pos -= center
                pos[:] = pos - np.rint(pos/side)*side
                pos += center
            else:
                pos[:] = pos - np.rint(pos/side)*side
            return self
        elif method == 'f90':
            _f90.fold(pos, self.system.cell.center, self.system.cell.side)

    def run(self, steps):
        """
        Evolve the system, as prescribed by `atooms.simulation.core.Simulation` class.  

        User defined orientation update methods can be provided via `self.neighbors` class variable. 
        """
        # Setup arrays (get views on atooms.system arrays)
        positions = self.system.view('pos', order='F')
        orientations = self.system.view('orientation', order='F')
        velocities = self.system.view('velocity', order='F')
        box = self.system.cell.side

        # Instantiate default neighbor list
        if self.neighbors == None:
            from neighbors import VicsekNeighbors
            self.neighbors = VicsekNeighbors(self.system, method='kdtree')

        # Main loop
        for i in range(steps):
            # Get sum of neighbors of each particle, its number of neighbors and timer data
            self.sum_neighbors, self.number_neighbors = self.neighbors.compute(box, positions, orientations)

            # Compute orientations according to noise kind specified by user
            orientations[:] = self.noise(orientations, self.number_neighbors,
                                         self.sum_neighbors, self.eta, len(self.system.particle))

            # Update velocities
            dv = np.asarray([self.v0*np.cos(orientations),
                             self.v0*np.sin(orientations)])
            velocities[:, :] = dv

            # Update positions
            dq = dv*self.deltat
            positions += dq

            # Apply PBC
            self._fold()



"""
Fixed radius nearest neighbors computation tools.
"""

import os
import numpy as np
from f2py_jit import jit
from atooms.core.utils import Timer

__all__ = ['VicsekNeighbors']


# Neighbors

class _Neighbors:
    def __init__(self, system, rcut=1., method='kdtree'):
        self.system = system
        self.rcut = rcut
        self.neighbors = None
        self.number_neighbors = None
        self.npart = len(self.system.particle)
        # Timers
        self.timer = {'id2ori': Timer(), 'f90': Timer(), 'kdtree': Timer()}

    def compute(self, box, positions, orientations):
        """
        Compute nearest neighbors in fixed radius.

        This function *must* return 
        - `self.sum_neighbors`: sum of neighbor orientation for each particle
        - `self.number_neighbors`: number of neighbors for each particle
        """
        pass

# Neighbors algorithms

class Python(_Neighbors):
    def __init__(self, system):
        _Neighbors.__init__(self, system, npart)

    def neighborold(self):
        """
        Legacy O(n*n) neighbor search algorithm

        Useful for unit tests; it isn't compatible with 'vicsek.py'
        """
        # TODO: adapt to standard compute method

        orients = []
        for o in self.system.particle:
            center = o.position

            orients.append([])

            for p in self.system.particle:  # append point to to numpy array
                if p is o:
                    continue
                dr = p.distance(o, self.system.cell)
                dr = np.sum(dr**2)**0.5
                if dr <= self.rcut:
                    orients[-1].append(float(p.orient))

        return orients

class F90(_Neighbors):
    def __init__(self, system, inline=False, f90file='neighbor_list_newton_inline.f90'):
        _Neighbors.__init__(self, system)

        self.f90file = f90file
        self.inline = inline
        self.ids = np.ones(self.npart)

    def _setup(self, nneigh):
        """
        Setup arrays for neighbor search in Fortran, using (estimated) number of neighbors for each particle
        """

        if self.neighbors is None or self.neighbors.shape[1] != self.npart or self.neighbors.shape[0] < nneigh:
            self.neighbors = np.ndarray(shape=(nneigh, self.npart),
                                        order='F', dtype=np.int32)
        if self.number_neighbors is None or len(self.number_neighbors) != self.npart:
            self.number_neighbors = np.ndarray(
                self.npart, order='F', dtype=np.int32)

    def _index_to_sum(self, orientations):
        """Convert neighbor indices to proper orientations"""

        # Setup arrays
        sum_neighbors = np.zeros(self.npart, dtype=complex)
        orientations = np.exp(1j*orientations)

        # Compute neighbor sum for each particle
        self.timer['id2ori'].start()
        for i in range(self.npart):
            j = self.number_neighbors[i]
            sum_neighbors[i] = np.sum(orientations[self.neighbors[0:j, i]].tolist())
        self.timer['id2ori'].stop()

        return sum_neighbors

    def compute(self, box, positions, orientations):
        """Compute nearest neighbors in radius using Fortran kernel"""

        # Initialize F90 module
        d = os.path.join(os.path.dirname(__file__), self.f90file)
        f90 = jit(d, inline=self.inline, flags="-O3 -ffast-math")

        # Initialize and setup arrays
        nneigh = int(np.pi * np.max(self.rcut)**2 * self.system.density * 1.)
        # nneigh : estimated number of neighs given starting density in circular area
        self._setup(nneigh)

        # Call Fortran function
        self.timer['f90'].start()
        err = f90.neighbor_list.compute(
            box, positions, self.ids, self.rcut, self.neighbors, self.number_neighbors)

        # Error flag handling
        if err:
            nneigh = max(self.number_neighbors)
            self._setup(nneigh)
            err = f90.neighbor_list.compute(
                box, positions, self.ids, self.rcut, self.neighbors, self.number_neighbors)
            assert not err, "something wrong with neighbor_list"
        self.timer['f90'].stop()

        # Shift Fortran ids to Python ids
        self.neighbors -= 1

        # Convert neighbors' indices to sum of neighbors' orientations
        self.sum_neighbors = self._index_to_sum(orientations)

        return self.sum_neighbors, self.number_neighbors

class KDTree(_Neighbors):
    def __init__(self, system):
        _Neighbors.__init__(self, system)

    def compute(self, box, positions, orientations):
        """
        Compute nearest neighbors in radius using kD-trees
        --------------------------------------------------
        kD-tree algorithm implementation: Francesco Turci and Daniele Coslovich
        """

        # Import needed pkgs
        from scipy import sparse
        from scipy.spatial import cKDTree

        # kD-tree algorithm
        self.timer['kdtree'].start()
        # Shift center to L/2 and transpose Fortran arrays to order='C'
        hbox = box/2.
        positions = np.transpose(positions) + hbox
        orient = np.transpose(orientations)
        # Define kD-tree and distance matrix
        tree = cKDTree(positions, boxsize=box)
        dist = tree.sparse_distance_matrix(tree, max_distance=self.rcut, output_type='coo_matrix')
        # important 3 lines: we evaluate a quantity for every column j
        data = np.exp(orient[dist.col]*1j)
        # construct a new sparse marix with entries in the same places ij of the dist matrix
        neighbors = sparse.coo_matrix((data, (dist.row, dist.col)), shape=dist.get_shape())
        # and sum along the columns (sum over j)
        sum_neighbors = np.squeeze(np.asarray(neighbors.tocsr().sum(axis=1)))
        # also get number of neighbors for each row/particle
        self.number_neighbors = neighbors.tocsr().getnnz(axis=0)
        self.timer['kdtree'].stop()

        return sum_neighbors, self.number_neighbors

# Interface

class VicsekNeighbors:
    """
    Interface for fixed-radius nearest neighbor search algorithms.

    Parameters
    ----------
    method: {'kdtree', 'f90'}
            Algorithm selection. Available algorithms are `KDTree`, `F90`, and `Python` (the latter to be used only for testing purposes).

    ## `F90` 
    Compute fixed-radius nearest neighbors using a O(n*n), Fortran algorithm. 
    The Fortran kernel is compiled just-in-time via `f2py-jit`; multiple choices are available

    - `neighbor_list.f90`: a basic implementation of the algorithm, without speed-ups, with inlining;
    - `neighbor_list_newton.f90`: Newton's third law is used to roughly halve computation times, keeping all other features intact;
    - `neighbor_list_inline.f90`: currently the fastest implementation, which doesn't rely on inlining (this might change in the future).

    Parameters
    ----------

    f90file:    string, optional
                Fortran kernel file name
    inline:     bool, default True
                Enable inlining in Fortran kernel execution. Currently, this is set to `False` due to performance issues.

    Callable using `method='f90'`.

    ## `KDTree` (default)
    Compute fixed-radius nearest neighbors using a O(n*log n), Python algorithm.
    This algorithm stems from [this](https://framagit.org/coslo/vicsek) implementation by Daniele Coslovich and Francesco Turci, and relies on `scipy.spatial.cKDTree`.

    Callable using `method='kdtree'`.

    ## `Python`
    Compute fixed-radius nearest neighbors using a basic, O(n*n), Python algorithm. 
    This mustn't be used outside testing. 

    It isn't callable via `method` variable. 
    """

    def __init__(self, system, method='kdtree', **kwargs):
        # Set noise
        if method == 'kdtree':
            self.method = KDTree(system)
        elif method == 'f90':
            self.method = F90(system, **kwargs)
        else:
            raise ValueError(f'wrong method {method}')

    def compute(self, box, positions, orientations):
        """
        Interface for `active.vicsek`.

        Parameters
        ----------

        box:            float or Numpy array
                        Simulation cell size.
        positions:      Numpy array
                        Particle positions. Must be Fortran ordered (`order='F'`).
        orientations:   Numpy array
                        Particle orientations. Must be Fortran ordered (`order='F'`).

        """
        return self.method.compute(box, positions, orientations)
